#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <termios.h>
#include <unistd.h>

#include <sys/io.h>
#include <sys/ioctl.h>
#include <linux/serial.h>
#include <linux/serial_reg.h>

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>


int verbose = 0;
int sendwait = 1000;
char *serial_dev = "/dev/ttyS0";
struct termios saved_ios, ios;

/*======================================================================
 */
void print_quoted(char *prefix, char *s, int len)
{
    int i;

    printf("%s<", prefix);
    for(i = 0; i < len; i++)
	if(s[i] < ' ') {
	    putchar('^');
	    putchar(s[i] + '@');
	} else {
	    putchar(s[i]);
	}
    putchar('>');
    putchar('\n');
}


int rs_send(int fd, char *s, int len)
{
    int n;

    if(len == 0)
	len = strlen(s);
    if(verbose)
	print_quoted("pc>mpc ", s, len);

    while(len > 0) {
	n = write(fd, s, 1);
	if(n < 0) {
	    perror("rs_send");
	    return n;
	}
	usleep(sendwait); /* some wait */

	len -= n;
	s += n;
    }
    return 0;
}


int rs_recv(int fd, char *s, int len)
{
    int n;

    n = read(fd, s, len);
    if(n < 0) {
	perror("rs_recv");
	return n;
    }
    if(verbose) {
	if(n == 0)
	    print_quoted("pc<mpc [EOF]", "", 0);
	else
	    print_quoted("pc<mpc ", s, n);
    }
    return n;
}


/*----------------------------------------------------------------------
 * expect():
 * repeat reading and searching specified byte sequence
 * return value:
 *  <=0: EOF or error or not found
 *  0>: found expected string
 */
int expect(int fd, char *expect, char *buf, int buf_max)
{
    char s[512];
    int s_len, buf_len;

    buf_len = 0;
    --buf_max; /* allocate room for terminating NUL */

    while(buf_len < buf_max) {
	if(buf_max - buf_len < sizeof(s))
	    s_len = rs_recv(fd, s, buf_max - buf_len);
	else
	    s_len = rs_recv(fd, s, sizeof(s));
	if(s_len <= 0) return -buf_len;

	memcpy(&buf[buf_len], s, s_len); /* append s to buf */
	buf_len += s_len;
	buf[buf_len] = 0;

	if(memcmp(buf + buf_len - strlen(expect), expect, strlen(expect)) == 0)
	    return buf_len; /* found */
    }
    return -buf_len;
}


/*======================================================================
 * synchronous transmission
 *
 * RxD/DSR = RD/DR  <- MISO
 * TxD (data/break) -> MOSI
 * DTR/CTS = ER     -> RESET
 * RTS     = RS     -> SCK
 */

#define HIGH		1
#define LOW		0

#define rs_get_miso	rs_get_dsr
#define rs_set_mosi	rs_set_break
#define rs_set_sck	rs_set_rts

__inline__ int rs_get_dsr(int port)
{
    return inb_p(port + UART_MSR) & UART_MSR_DSR;
}

__inline__ void rs_set_break(int port, int flag)
{
    if(flag)
	outb_p(inb_p(port + UART_LCR) | UART_LCR_SBC, port + UART_LCR);
    else
	outb_p(inb_p(port + UART_LCR) & ~UART_LCR_SBC, port + UART_LCR);
}

__inline__ void rs_set_rts(int port, int flag)
{
    if(flag)
	outb_p(inb_p(port + UART_MCR) | UART_MCR_RTS, port + UART_MCR);
    else
	outb_p(inb_p(port + UART_MCR) & ~UART_MCR_RTS, port + UART_MCR);
}


/*----------------------------------------------------------------------
 * send byte synchronously
 */
void sync_send(int port, int bytedata)
{
    rs_set_sck(port, LOW);
    rs_set_mosi(port, bytedata & 0x01);
    rs_set_sck(port, HIGH);
    rs_set_mosi(port, bytedata & 0x02);
    rs_set_sck(port, LOW);
    rs_set_mosi(port, bytedata & 0x04);
    rs_set_sck(port, HIGH);
    rs_set_mosi(port, bytedata & 0x08);
    rs_set_sck(port, LOW);
    rs_set_mosi(port, bytedata & 0x10);
    rs_set_sck(port, HIGH);
    rs_set_mosi(port, bytedata & 0x20);
    rs_set_sck(port, LOW);
    rs_set_mosi(port, bytedata & 0x40);
    rs_set_sck(port, HIGH);
    rs_set_mosi(port, bytedata & 0x80);
    rs_set_sck(port, LOW);
}

/*----------------------------------------------------------------------
 * send MP3 data synchronously
 */
#define XFER_BLOCK	512

int rs_send_sync(int fd, char *path)
{
    int i, j, nblk;
    char buf[XFER_BLOCK];
    struct serial_struct si;
    FILE *fp;
    char *ack = "\006", *eot = "\004";

    /* open MP3 file and get size */
    printf("info: sending MP3 file: %s\n", path);
    if((fp = fopen(path, "rb")) == NULL) {
	fprintf(stderr, "rs_send_sync: %s: can't open\n", path);
	return -1;
    }
    for(nblk = 0; !feof(fp); nblk++)
	fread(buf, XFER_BLOCK, 1, fp);

    /* get I/O address of serial */
    ioctl(fd, TIOCGSERIAL, &si);
    if(verbose) {
	printf("info: MP3 file is %d blocks\n", nblk);
	printf("info: port 0x%x irq 0x%x\n", si.port, si.irq);
    }

    rs_send(fd, "W\r", 0);
    expect(fd, "Please send MPEG data..\r\n", buf, sizeof(buf));

    /* set SCK=LOW, and send # of blocks */
    rs_set_sck(si.port, LOW);
    buf[0] = (char)nblk;
    buf[1] = (char)(nblk >> 8);
    buf[2] = (char)(nblk >> 16);
    rs_send(fd, buf, 3);

    /* transfer blocks */
    fseek(fp, 0, SEEK_SET);
    for(i = 0; i < nblk; i++) {
	if(verbose)
	    fprintf(stderr, "info: sending %d of %d...\r", i + 1, nblk);

	expect(fd, ack, buf, sizeof(buf));
	fread(buf, XFER_BLOCK, 1, fp);
	for(j = 0; j < XFER_BLOCK; j++)
	    sync_send(si.port, buf[j]);
	tcdrain(fd); /* flush any pending transmission */
    }
    expect(fd, eot, buf, sizeof(buf));

    /* set SCK=HIGH */
    rs_set_sck(si.port, HIGH);

    if(verbose)
	fprintf(stderr, "info: sending %d blocks done\n", nblk);
    return 0;
}

void rs_send_mp3_data(int fd, char **paths)
{
    char **p;

    for(p = paths; *p; p++)
	rs_send_sync(fd, *p);
    printf("MP3 data transmission done\n");
}

/*======================================================================*/

void show_mp3_list(char *s, int n)
{
    /* skip "L\r\n" */
    s += 3; n -= 3;
    /* skip "\r\n% " */
    n -= 4;
    while(n > 0) {
	if(*s == 0) putchar(' ');
	else if(*s != '\r') putchar(*s);
	++s;
	--n;
    }
    putchar('\n');
}


/*----------------------------------------------------------------------
 */
void print_help(void)
{
    char **p, *msgs[] = {
	"elmmpc " VERSION " - ELM-MPC MP3 player utility for Linux", 
	"",
	"Usage: mpc [-f file] [-v] OPERATION [ARGS...]"
	"",
	"Operations:",
	"  list",
	"  play N",
	"  delete N",
	"  deleteall",
	"  write FILE",
	"  renum N M",
	"  format-media",
	"",
	0
    };

    for(p = msgs; *p; p++)
	printf("%s\n", *p);
}


/*----------------------------------------------------------------------
 */
int main(int argc, char *argv[])
{
    int argi, fd, n;
    char buf[4096];
    char *prompt = "\r\n% ";

    /*
     * drop superuser's privilege except I/O access
     */
    if(iopl(3) < 0) {
	perror("iopl");
	return 1;
    }
    if(setuid(getuid()) < 0) {
	perror("setuid");
	return 1;
    }

    /*
     * analyze common options
     */
    for(argi = 1; argi < argc; argi++) {
	if(strcmp(argv[argi], "-f") == 0 && (argi + 1 < argc)) {
	    serial_dev = argv[argi + 1];
	    ++argi;
	    continue;
	}

	if(strcmp(argv[argi], "-v") == 0) {
	    verbose = 1;
	    continue;
	}

	break;
    }

    /*
     * sets up serial communication
     */
    fd = open(serial_dev, O_RDWR);
    if(fd < 0) {
	perror("open");
	return 1;
    }
    tcgetattr(fd, &saved_ios);
    ios = saved_ios;
    cfmakeraw(&ios);
    cfsetspeed(&ios, B38400);
    tcsetattr(fd, TCSANOW, &ios);

    if(expect(fd, prompt, buf, sizeof(buf)) <= 0) {
	fprintf(stderr, "error: unexpected response from ELM-MPC\n");
	return 1;
    }
    if(verbose)
	fprintf(stderr, "info: connected to ELM-MPC\n");

    /*
     * process specified operation
     */
    if(argi >= argc) {
	fprintf(stderr, "no opearation supplied\n");
	return 1;
    }

    if(strcmp(argv[argi], "list") == 0
	      || strcmp(argv[argi], "ls") == 0) {
	rs_send(fd, "L\r", 0);
	n = expect(fd, prompt, buf, sizeof(buf));
	show_mp3_list(buf, n);
    } else if(strcmp(argv[argi], "play") == 0 && (argi + 1 < argc)) {
	sprintf(buf, "P%d\r", atoi(argv[argi + 1]));
	rs_send(fd, buf, 0);
	expect(fd, prompt, buf, sizeof(buf));
    } else if((strcmp(argv[argi], "delete") == 0
	       || strcmp(argv[argi], "rm") == 0) && (argi + 1 < argc)) {
	sprintf(buf, "D%d\r", atoi(argv[argi + 1]));
	rs_send(fd, buf, 0);
	expect(fd, prompt, buf, sizeof(buf));
    } else if(strcmp(argv[argi], "deleteall") == 0) {
	rs_send(fd, "DA\r", 0);
	expect(fd, prompt, buf, sizeof(buf));
    } else if(strcmp(argv[argi], "renum") == 0 && (argi + 2 < argc)) {
	sprintf(buf, "R%d %d\r", atoi(argv[argi + 1]), atoi(argv[argi + 2]));
	rs_send(fd, buf, 0);
	expect(fd, prompt, buf, sizeof(buf));
    } else if(strcmp(argv[argi], "format-media") == 0) {
	rs_send(fd, "FM\r", 0);
	expect(fd, prompt, buf, sizeof(buf));
	printf("formatting of media done\n");
    } else if(strcmp(argv[argi], "write") == 0 && (argi + 1 < argc)) {
	rs_send_mp3_data(fd, &argv[argi + 1]);
    } else {
	print_help();
    }

    /*
     * closes communication
     */
    tcsetattr(fd, TCSANOW, &saved_ios);
    close(fd);
    return 0;
}
